;; Pinaret - Static website generator
;; Copyright (C) 2016  Josep Portella Florit <hello@josep-portella.com>
;;
;; This program is free software: you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation, either version 3 of the License, or
;; (at your option) any later version.
;;
;; This program is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.

#lang racket

(require "provide.rkt" "runtime.rkt" "procedure.rkt")

(define/provide key-char (make-parameter #\|))
(define/provide comment-key (make-parameter '|#|))

(define/provide (parse-dict proc init input)
  (define (skip-whitespace-and-pipe)
    (let ([c (peek-char input)])
      (when (and (not (eof-object? c))
                 (or (char-whitespace? c)
                     (eqv? c (key-char))))
        (read-char input)
        (skip-whitespace-and-pipe))))
  (define (read-key)
    (skip-whitespace-and-pipe)
    (let-values ([(line column _) (port-next-location input)])
      (values (call-with-output-string
               (lambda (output)
                 (let loop ([c (peek-char input)])
                   (define (after-newline)
                     (unless (eqv? (peek-char input 1) (key-char))
                       (read-char input)))
                   (cond [(eof-object? c)]
                         [(eqv? c #\return)
                          (when (eqv? (peek-char input 1) #\newline)
                            (read-char input))
                          (after-newline)]
                         [(eqv? c #\newline) (after-newline)]
                         [(char-whitespace? c) (read-char input)]
                         [else (write-char (read-char input) output)
                               (loop (peek-char input))]))))
              line column)))
  (define (read-value)
    (let-values ([(line column _) (port-next-location input)])
      (values
       (call-with-output-string
        (lambda (output)
          (let loop ([c (read-char input)])
            (define (after-newline x)
              (let ([d (read-char input)])
                (unless (or (eof-object? d) (eqv? d (key-char)))
                  (display x output)
                  (if (eqv? d (escape-char))
                      (let ([e (read-char input)])
                        (cond [(eqv? e (key-char))
                               (write-char (key-char) output)
                               (loop (read-char input))]
                              [else (write-char (escape-char) output)
                                    (loop (if (eqv? (escape-char) e)
                                              (read-char input)
                                              e))]))
                      (loop d)))))
            (cond [(eof-object? c)]
                  [(eqv? c #\return)
                   (cond [(eqv? (peek-char input) #\newline)
                          (read-char input)
                          (after-newline "\r\n")]
                         [else (after-newline #\return)])]
                  [(eqv? c #\newline) (after-newline #\newline)]
                  [else (write-char c output)
                        (loop (read-char input))]))))
       line column)))
  (when (eqv? (peek-char input) #\uFEFF) (read-char input))
  (port-count-lines! input)
  (let loop ([state init])
    (call-with-values read-key
      (lambda (key kline kcol)
        (if (equal? "" key)
            state
            (call-with-values read-value
              (lambda (value vline vcol)
                (loop (proc key kline kcol value vline vcol state)))))))))

(define/provide (read-dict-file path [hsh (hasheq)] #:thunk-maker->procedure
                                [t-m->p thunk-maker->thunk])
  (with-handlers ([exn:fail? (error/params-raiser)])
    (let ([port (open-input-file path)])
      (dynamic-wind
        void
        (lambda ()
          (with-handlers
              ([exn:fail? (error/params-raiser 'read-dict-file path)])
            (parse-dict
             (lambda (key kline kcol value vline vcol hsh*)
               (let ([sym (string->symbol key)])
                 (if (eq? (comment-key) sym)
                     hsh*
                     (let ([proc (t-m->p (string->thunk-maker
                                          value path vline vcol))]
                           [pos (position path kline kcol)])
                       (cond [(hash-ref hsh* sym #f) =>
                              (lambda (proc*)
                                (raise-error/params
                                 'multiple-definitions sym
                                 'defined-in pos
                                 'defined-in (procedure-position proc*)))]
                             [else (hash-set hsh* sym (procedure/properties
                                                       proc pos sym))])))))
             hsh port)))
        (λ () (close-input-port port))))))
