;; Pinaret - Static website generator
;; Copyright (C) 2016  Josep Portella Florit <hello@josep-portella.com>
;;
;; This program is free software: you can redistribute it and/or modify
;; it under the terms of the GNU General Public License as published by
;; the Free Software Foundation, either version 3 of the License, or
;; (at your option) any later version.
;;
;; This program is distributed in the hope that it will be useful,
;; but WITHOUT ANY WARRANTY; without even the implied warranty of
;; MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
;; GNU General Public License for more details.
;;
;; You should have received a copy of the GNU General Public License
;; along with this program.  If not, see <http://www.gnu.org/licenses/>.

#lang racket

(require "provide.rkt" "runtime.rkt" "procedure.rkt")

(define/provide repl-prompt (make-parameter "\n> "))
(define/provide repl-prompt* (make-parameter "| "))

(define (prompt str)
  (display str)
  (read-line))

(define (continue-reading? e)
  (member (first (error/params-args e))
          '(unfinished-quoting
            unfinished-escape-sequence
            unfinished-procedure-application)))

(define/provide (repl [init-parameters void])
  (let loop ([line (prompt (repl-prompt))])
    (unless (eof-object? line)
      (with-handlers ([error/params? (lambda (e)
                                       (print-error/params e)
                                       (loop (prompt (repl-prompt))))])
        (init-parameters)
        (loop (call/cc
               (lambda (continue)
                 (display ((with-handlers
                               ([error/params?
                                 (lambda (e)
                                   (if (continue-reading? e)
                                       (continue
                                        (let ([line*
                                               (prompt (repl-prompt*))])
                                          (cond [(eof-object? line*)
                                                 (print-error/params e)
                                                 line*]
                                                [else (string-append
                                                       line "\n" line*)])))
                                       (raise e)))])
                             (thunk-maker->thunk
                              (string->thunk-maker line)))))
                 (prompt (repl-prompt)))))))))
